/*******************************************************************************
#	 	luvcview: Sdl video Usb Video Class grabber           .        #
#This package work with the Logitech UVC based webcams with the mjpeg feature. #
#All the decoding is in user space with the embedded jpeg decoder              #
#.                                                                             #
# 		Copyright (C) 2005 2006 Laurent Pinchart &&  Michel Xhaard     #
#                                                                              #
# This program is free software; you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by         #
# the Free Software Foundation; either version 2 of the License, or            #
# (at your option) any later version.                                          #
#                                                                              #
# This program is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of               #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                #
# GNU General Public License for more details.                                 #
#                                                                              #
# You should have received a copy of the GNU General Public License            #
# along with this program; if not, write to the Free Software                  #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    #
#                                                                              #
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <string.h>
#include <pthread.h>
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include "v4l2uvc.h"

#define INCPANTILT 64 // 1
static const char version[] = VERSION;
struct vdIn *videoIn;
    
int main(int argc, char *argv[])
{
    const char *videodevice = NULL;
    int format = V4L2_PIX_FMT_MJPEG;
    int i;
    int grabmethod = 1;
    int width = 320;
    int height = 240;
    int fps = 15;
    int queryformats = 0;
    int querycontrols = 0;

    printf("luvcview version %s \n", version);
    for (i = 1; i < argc; i++) {
		
		/* skip bad arguments */
		if (argv[i] == NULL || *argv[i] == 0 || *argv[i] != '-') {
			continue;
		}
		
		if (strcmp(argv[i], "-d") == 0) {
			if (i + 1 >= argc) {
				printf("No parameter specified with -d, aborting.\n");
				exit(1);
			}
			videodevice = strdup(argv[i + 1]);
		}
		
		if (strcmp(argv[i], "-L") == 0) {
			/* query list of valid video formats */
			queryformats = 1;
		}
		
		if (strcmp(argv[i], "-l") == 0) {
			/* query list of valid video formats */
			querycontrols = 1;
		}
		
		if (strcmp(argv[i], "-h") == 0) {
			printf("usage: uvcview [-h -d -L -l] \n");
			printf("-h	print this message \n");
			printf("-d	/dev/videoX       use videoX device\n");
			printf("-L	query valid video formats\n");
			printf("-l	query valid controls and settings\n");
			exit(0);
		}
		
    }

    if (videodevice == NULL || *videodevice == 0) {
		videodevice = "/dev/video0";
    }

    videoIn = (struct vdIn *) calloc(1, sizeof(struct vdIn));
	
     /* if we're supposed to list the video formats, do that now and go out */
    if ( queryformats ) {
     	check_videoIn(videoIn,(char *) videodevice);
    	free(videoIn);
		exit(1);
	}
	
	/* if we're supposed to list the controls, do that now */
    if ( querycontrols )
	{
		if (init_videoIn(videoIn, (char *) videodevice, width, height, fps, format, grabmethod) < 0)
			exit(1);
        enum_controls(videoIn->fd);
	}
	
    close_v4l2(videoIn);
    free(videoIn);
    printf(" Clean Up done Quit \n");
	
	return 0;
}
